/*=============================================================================
	Module Name:
			m2audio_ac3.cpp

	Abstract:
			m2Audio AC-3 decoder plugin for AviSynth 2.5

	License:
            GNU General Public License (GPL)

	Revision History:
			* Mar 29, 2004:  Created by Attila Afra
			* Jul  5, 2004:  DAI (DVD Audio Index) input
			* Jul 20, 2004:  Major bugfixes, improved error tolerance
			* Sep 12, 2004:  DAI replaced with VFM
			* Nov  6, 2004:  Bugfixes
			* Feb 13, 2005:  Minor changes
			* Feb 27, 2005:  Advanced dummy audio generation

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

#include <windows.h>
#include <string.h>
#include "vfm.h"
#include "avisynth.h"

#include "AudioHeader.h"

/*-----------------------------------------------------------------------------
	Constants
-----------------------------------------------------------------------------*/

#define FRAME_HEADER_LEN	7										// length of a frame's header
#define BLOCKS_PER_FRAME	6										// number of blocks in a frame
#define SAMPLES_PER_BLOCK	256										// number of samples per channel in a block
#define SAMPLES_PER_FRAME	(BLOCKS_PER_FRAME * SAMPLES_PER_BLOCK)	// number of samples per channel in a frame

#define MAX_DETECT_ERROR	1024									// max detection error (in bytes)

/*-----------------------------------------------------------------------------
	m2AudioAC3Source: Input filter class
-----------------------------------------------------------------------------*/


/*=============================================================================
	m2AudioAC3Source >> Implementation
=============================================================================*/

AVSValue __cdecl m2AudioAC3Source::Create(AVSValue args, void *, IScriptEnvironment *env)
{
	return new m2AudioAC3Source(args[0].AsString(), args[1].AsInt(0), args[2].AsInt(0), env);
}

/*-----------------------------------------------------------------------------
	m2AudioAC3Source >> Initialization
-----------------------------------------------------------------------------*/

// Constructor
m2AudioAC3Source::m2AudioAC3Source(const char *FileName, int Downmix, int DRC, IScriptEnvironment *env)
{
	int Channels;	// channel count
	int Sync;		// synch info
	int _DecFlags;	// temporary flags
	
	// Initialize pointers
	Stream = NULL;
	State  = 0;
	Frame  = 0;
	Buffer = 0;
	_drc = DRC;
	
	// Store filename
	StreamName = new char[strlen(FileName) + 1];
	strcpy(StreamName, FileName);
	
	// Open file
	if (!(Stream = vfopen(StreamName)))
		env->ThrowError("m2AudioAC3Source: unable to open file \"%s\"", StreamName);

	StreamLength = vflen(Stream);

	// Empty information structure
	memset(&Info, 0, sizeof(VideoInfo));

	// Check if the stream is empty
	if (!StreamLength)
	{
		// Dummy audio
		Info.sample_type              = SAMPLE_INT16;
		Info.nchannels                = Downmix ? Downmix : 2;
		Info.audio_samples_per_second = 44100;
		Info.num_audio_samples        = 0;

		SampleCount                   = 0;

		return;
	}

	// Find first frame
	FrameLength = MAX_DETECT_ERROR;
	Sync        = Synchronize(FrameLength, Flags, Samplerate, Bitrate);
	if (Sync <= 0)
		env->ThrowError("m2AudioAC3Source: \"%s\" is not a valid AC-3 file", StreamName);

	StreamOffset = (Sync - 1) % FrameLength;
	vfseek(Stream, StreamOffset, SEEK_SET);

	// Set sample type
	Info.sample_type = SAMPLE_FLOAT;

	// Get channel count
	LFE      = false;
	Channels = Flags & A52_CHANNEL_MASK;

	switch (Channels)
	{
	case A52_MONO:
		DecFlags     = A52_MONO;
		ChannelCount = 1;
		break;

	case A52_STEREO:
		DecFlags     = A52_STEREO;
		ChannelCount = 2;
		break;

	case A52_DOLBY:
	case A52_3F:
	case A52_2F1R:
	case A52_3F1R:
		DecFlags     = A52_DOLBY;
		ChannelCount = 2;
		break;

	case A52_2F2R:
		DecFlags     = A52_2F2R;
		ChannelCount = 4;
		break;

	case A52_3F2R:
		LFE      = (Flags & A52_LFE) == A52_LFE; // WAS Flags && A52_LFE !!!!!
		DecFlags = A52_3F2R;
		if (LFE)
			DecFlags |= A52_LFE;
		ChannelCount = 6;
		break;

	default:
		env->ThrowError("m2AudioAC3Source: unsupported channel configuration in file \"%s\"", StreamName);
	}

	if (Downmix)
	{
		// Downmix
		switch (Downmix)
		{
		// 1 channel
		case 1:
			_DecFlags = A52_MONO;
			break;

		// 2 channels (Dolby Surround)
		case 2:
			_DecFlags = A52_DOLBY;
			break;

		// 4 channels
		case 4:
			_DecFlags = A52_2F2R;
			break;

		// 6 channels
		case 6:
			_DecFlags = A52_3F2R;
			if (LFE)
				_DecFlags |= A52_LFE;
			break;

		default:
			env->ThrowError("m2AudioAC3Source: invalid number of channels requested");
		}

		if (Downmix < ChannelCount)
		{
			ChannelCount = Downmix;
			DecFlags     = _DecFlags;
		}
	}

	Info.nchannels = ChannelCount;

	// Enable automatic level adjustment
	DecFlags |= A52_ADJUST_LEVEL;
	Level     = 1;

	// Get sampling rate
	Info.audio_samples_per_second = Samplerate;

	// Get number of frames
	FrameCount = int((StreamLength - StreamOffset) / FrameLength);
	
	// Get number of samples
	SampleCount            = __int64(FrameCount) * SAMPLES_PER_FRAME;
	Info.num_audio_samples = SampleCount;

	// Initialize buffers
	Frame	    = new unsigned char[FrameLength];
	FrameIndex  = -2;

	BufferSize  = Info.nchannels * SAMPLES_PER_FRAME;
	Buffer      = new float[BufferSize];
	BufferStart = BufferEnd = -1;

	// Set decoder acceleration
	Accel = 0;
}

// Destructor
m2AudioAC3Source::~m2AudioAC3Source()
{
	// Close file
	if (Stream)
		vfclose(Stream);

	// Free memory
	if (State)
		a52_free(State);

	if (Frame)
		delete[] Frame;

	if (Buffer)
		delete[] Buffer;
}

/*-----------------------------------------------------------------------------
	m2AudioAC3Source >> IClip interface
-----------------------------------------------------------------------------*/

// Return VideoInfo
const VideoInfo& __stdcall m2AudioAC3Source::GetVideoInfo()
{
	return Info;
}

// Return video frame
PVideoFrame __stdcall m2AudioAC3Source::GetFrame(int n, IScriptEnvironment *env)
{
	return NULL;
}

// Return video field parity
bool __stdcall m2AudioAC3Source::GetParity(int n)
{
	return false;
}

// Return audio samples
void __stdcall m2AudioAC3Source::GetAudio(void *buf, __int64 start, __int64 count, IScriptEnvironment *env)
{
	__int64   i;				// cycle counter
	int      _FrameIndex;		// index of the previous frame
	float   *Input, *Output;	// buffer pointers
	__int64  Left;			    // number of samples to be copied from the buffer

	// Check if the requested samples are not out of range
	if (!SampleCount || (start < 0) || (count < 1))
		return;

	if (start + count > SampleCount)
		count = SampleCount - start;
	
	// Check if we must read a frame
	if (start < BufferStart || start > BufferEnd)
	{
		// Calculate the index and the sample range of the required frame
		_FrameIndex = FrameIndex;
		FrameIndex  = int(start / SAMPLES_PER_FRAME);
		BufferStart = __int64(FrameIndex) * SAMPLES_PER_FRAME;
		BufferEnd   = BufferStart + SAMPLES_PER_FRAME;

		// Seek to the frame, if required
		if (FrameIndex != _FrameIndex + 1)
		{
			vfseek(Stream, StreamOffset + __int64(FrameIndex) * __int64(FrameLength), SEEK_SET);

			// Reset decoder
			if (State)
				a52_free(State);
			State = a52_init(Accel);
		}

		// Read the frame
		if (!ReadFrame())
			env->ThrowError("m2AudioAC3Source: error in file \"%s\"", StreamName);
	}

	// Start decoding
	Input  = Buffer + int(start - BufferStart) * ChannelCount;
	Output = (float *)buf;
	Left   = SAMPLES_PER_FRAME - (start - BufferStart);

	while (count)
	{
		// Copy samples
		if (count < Left)
			Left = count;
		count -= Left;

		for (i = 0; i < Left * ChannelCount; i++)
			*Output++ = *Input++;

		// Check if we must read a frame
		if (count)
		{
			// Read the frame
			FrameIndex++;
			if (!ReadFrame())
				env->ThrowError("m2AudioAC3Source: error in file \"%s\"", StreamName);

			// Reset input buffer pointer
			Input = Buffer;
			Left  = SAMPLES_PER_FRAME;
		}
	}

	// Calculate the sample range of the current frame
	BufferStart = __int64(FrameIndex) * SAMPLES_PER_FRAME;
	BufferEnd   = BufferStart + SAMPLES_PER_FRAME;
}

// Set cache hints
void __stdcall m2AudioAC3Source::SetCacheHints(int cachehints, int frame_range)
{
}

/*-----------------------------------------------------------------------------
	m2AudioAC3Source >> Operations
-----------------------------------------------------------------------------*/

// Synchronize with the audio stream
int m2AudioAC3Source::Synchronize(int& Length, int& Flags, int& Samplerate, int& Bitrate)
{
	int Delta = 0;								// number of bytes out of synch
	int _Length;								// backup of Length
	unsigned char Buffer[FRAME_HEADER_LEN];		// buffer for a frameheader

	// Backup Length
	_Length = Length;

	// Try synchronization
	while (Delta < _Length - 1)
	{
		// Read header
		if (vfread(Stream, Buffer, FRAME_HEADER_LEN) != FRAME_HEADER_LEN)
			return -1;

		// Get information
		Length = a52_syncinfo(Buffer, &Flags, &Samplerate, &Bitrate);

		if (Length)
		{
			// Success
			vfseek(Stream, -FRAME_HEADER_LEN, SEEK_CUR);
			return (Delta + 1);
		}
		else
		{
			// Try again
			vfseek(Stream, 1 - FRAME_HEADER_LEN, SEEK_CUR);
			Delta++;
		}
	}

	// Failure
	Length = _Length;
	return 0;
}

// Reads the next frame from the stream and decodes it into the framebuffer
bool m2AudioAC3Source::ReadFrame()
{
	int Sync;							// synch info
	int Length;							// length of the current frame
	int _Flags, _Samplerate, _Bitrate;	// bit stream information

	// Read next frame from stream
	if (vfread(Stream, Frame, FrameLength) != FrameLength)
	{
		// Error at the end of the stream, mute frame
		EmptyFrame();
		return true;
	}

	// Get frame information
	Length = a52_syncinfo(Frame, &_Flags, &_Samplerate, &_Bitrate);

	// Check if we need synchronization
	if (!Length)
	{
		// Seek back in the stream
		vfseek(Stream, 1 - FrameLength, SEEK_CUR);

		// Try synchronization
		Length = FrameLength;
		Sync   = Synchronize(Length, _Flags, _Samplerate, _Bitrate);
		
		// Serious damage in the stream, mute frame
		if (!Sync || Sync < 0)
		{
			// Mute frame
			EmptyFrame();
			return true;
		}

		// Check if the frame has proper length and BSI
		if ((Length != FrameLength) || (_Flags != Flags) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

		// Read frame
		if (vfread(Stream, Frame, FrameLength) != FrameLength)
		{
			// Error at the end of the stream, mute frame
			EmptyFrame();
			return true;
		}
	}
	else
		// Check if the frame has proper length and BSI
		if ((Length != FrameLength) || (_Flags != Flags) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

	// Decode frame
	if (!DecodeFrame())
		EmptyFrame();

	return true;
}

// Decode frame from the raw buffer
bool m2AudioAC3Source::DecodeFrame()
{
	int    blk, i;			// cycle counters
	float *Input, *Output;	// buffer pointers
	
	// Start decoding the frame
	if (a52_frame(State, Frame, &DecFlags, &Level, 0))
		return false;

	// Dynamic range compression !! dimzon !!

	if(0==_drc)
		a52_dynrng(State, 0, 0); 

	// The decoded samples will be in the following sequencial order:
	//   LFE, front left, front center, front right, rear left, rear right
	// If one of the channels is not present, it is skipped and the following
	// channels are shifted accordingly. We have to put the channels in
	// the proper interleaved order, based on the requested channel config.
	
	switch (ChannelCount)
	{
	// 1 channel
	case 1:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BLOCKS_PER_FRAME; blk++)
		{
			// Decode block
			if (a52_block(State))
				return false;

			// Copy channel
			Input = a52_samples(State);

			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				*Output++ = *Input++;
		}
		break;

	// 2 channels: left, right
	case 2:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BLOCKS_PER_FRAME; blk++)
		{
			// Decode block
			if (a52_block(State))
				return false;

			// Interleave channels
			Input = a52_samples(State);

			// left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
		break;

	// 4 channels: front left, front right, rear left, rear right
	case 4:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BLOCKS_PER_FRAME; blk++)
		{
			// Decode block
			if (a52_block(State))
				return false;

			// Interleave channels
			Input = a52_samples(State);

			// front left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// front right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// rear left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 2] = *Input++;

			// rear right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 3] = *Input++;

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
		break;

	// 6 channels: front left, front right, front center, LFE, rear left, rear right
	case 6:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BLOCKS_PER_FRAME; blk++)
		{
			// Decode block
			if (a52_block(State))
				return false;

			// Interleave channels
			Input = a52_samples(State);

			// LFE
			if (LFE)
			{
				for (i = 0; i < SAMPLES_PER_BLOCK; i++)
					Output[i * ChannelCount + 3] = *Input++;
			}
			else
			{
				for (i = 0; i < SAMPLES_PER_BLOCK; i++)
					Output[i * ChannelCount + 3] = 0;
			}
	
			// front left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// front center
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 2] = *Input++;

			// front right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// rear left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 4] = *Input++;

			// rear right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 5] = *Input++;

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
	}

	return true;
}

// Empty the framebuffer
void m2AudioAC3Source::EmptyFrame()
{
	for (int i = 0; i < SAMPLES_PER_FRAME * ChannelCount; i++)
		Buffer[i] = 0;
}

/*-----------------------------------------------------------------------------
	AviSynth plugin initialization
-----------------------------------------------------------------------------*/

/*
extern "C" __declspec(dllexport) const char * __stdcall AvisynthPluginInit2(IScriptEnvironment *env)
{
	env->AddFunction("NicAC3Source", "s[channels]i[drc]i", m2AudioAC3Source::Create, 0);
	env->AddFunction("NicDTSSource", "s[channels]i[drc]i", m2AudioDTSSource::Create, 0);
	env->AddFunction("NicLPCMSource", "s[samplerate]i[samplebits]i[channels]i", m2AudioLPCMSource::Create, 0);
	env->AddFunction("NicMPASource", "s[channels]i", m2AudioMPASource::Create, 0);

	return "Audio Decoder Plugin";
}
*/