/*=============================================================================
	Module Name:
			m2audio_dts.cpp

	Abstract:
			m2Audio DTS decoder plugin for AviSynth 2.5

	License:
            GNU General Public License (GPL)

	Revision History:
			* Apr 16, 2004:  Created by Attila Afra
			* Jul  9, 2004:  DAI (DVD Audio Index) input
			* Jul 20, 2004:  Major bugfixes, improved error tolerance
			* Sep 12, 2004:  DAI replaced with VFM
			* Nov  6, 2004:  Bugfixes
			* Feb 13, 2005:  Minor changes
			* Feb 27, 2005:  Advanced dummy audio generation

	Notes:
			* The termination frame is not decoded.

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

#include <windows.h>
#include <string.h>
#include "vfm.h"
#include "avisynth.h"

#include "AudioHeader.h"

#define FRAME_HEADER_LEN	14			// length of a frame's header

/*=============================================================================
	m2AudioDTSSource >> Implementation
=============================================================================*/

AVSValue __cdecl m2AudioDTSSource::Create(AVSValue args, void *, IScriptEnvironment *env)
{
	return new m2AudioDTSSource(args[0].AsString(), args[1].AsInt(0), args[2].AsInt(0), env);
}

/*-----------------------------------------------------------------------------
	m2AudioDTSSource >> Initialization
-----------------------------------------------------------------------------*/

// Constructor
m2AudioDTSSource::m2AudioDTSSource(const char *FileName, int Downmix, int DRC, IScriptEnvironment *env)
{
	int Channels;	// channel count
	int Sync;		// synch info
	int _DecFlags;	// temporary flags
	
	// Initialize pointers
	Stream = NULL;
	State  = 0;
	Frame  = 0;
	Buffer = 0;
	_drc = DRC;
	
	// Store filename
	StreamName = new char[strlen(FileName) + 1];
	strcpy(StreamName, FileName);
	
	// Open file
	if (!(Stream = vfopen(StreamName)))
		env->ThrowError("m2AudioDTSSource: unable to open file \"%s\"", StreamName);

	StreamLength = vflen(Stream);

	// Check if the stream is empty
	if (!StreamLength)
	{
		// Dummy audio
		Info.sample_type              = SAMPLE_INT16;
		Info.nchannels                = Downmix ? Downmix : 2;
		Info.audio_samples_per_second = 44100;
		Info.num_audio_samples        = 0;

		SampleCount                   = 0;

		return;
	}

	// Empty information structure
	memset(&Info, 0, sizeof(VideoInfo));

	// Set decoder acceleration
	Accel = 0;

	// Find first frame
	State       = dts_init(Accel);
	FrameLength = MAX_DETECT_ERROR;
	Sync = Synchronize(FrameLength, Flags, Samplerate, Bitrate);
	if (Sync <= 0)
		env->ThrowError("m2AudioDTSSource: \"%s\" is not a valid DTS file", StreamName);

	StreamOffset = (Sync - 1) % FrameLength;
	vfseek(Stream, StreamOffset, SEEK_SET);

	// Initialize raw buffer
	Frame      = new unsigned char[FrameLength];
	FrameIndex = -2;
	
	// Get number of blocks per frame
	vfread(Stream, Frame, FrameLength);
	vfseek(Stream, StreamOffset, SEEK_SET);
	if (dts_frame(State, Frame, &Flags, &Level, 0))
		env->ThrowError("m2AudioDTSSource: error in file \"%s\"", StreamName);

	BlocksPerFrame  = dts_blocks_num(State);
	SamplesPerFrame = BlocksPerFrame * SAMPLES_PER_BLOCK;
	dts_free(State);
	State = 0;

	// Set sample type
	Info.sample_type = SAMPLE_FLOAT;

	// Get channel count
	LFE      = false;
	Channels = Flags & DTS_CHANNEL_MASK;

	switch (Channels)
	{
	case DTS_MONO:
		DecFlags     = DTS_MONO;
		ChannelCount = 1;
		break;

	case DTS_STEREO:
	case DTS_3F:
	case DTS_2F1R:
	case DTS_3F1R:
		DecFlags     = DTS_STEREO;
		ChannelCount = 2;
		break;

	case DTS_DOLBY:
		DecFlags     = DTS_DOLBY;
		ChannelCount = 2;
		break;

	case DTS_2F2R:
		DecFlags     = DTS_2F2R;
		ChannelCount = 4;
		break;

	case DTS_3F2R:
		LFE = (0!=(Flags & DTS_LFE)); // WAZ Flags && DTS_LFE !!!!
		DecFlags = DTS_3F2R;
		if (LFE)
			DecFlags |= DTS_LFE;
		ChannelCount = 6;
		break;

	default:
		env->ThrowError("m2AudioDTSSource: unsupported channel configuration in file \"%s\"", StreamName);
	}

	if (Downmix)
	{
		// Downmix
		switch (Downmix)
		{
		// 1 channel
		case 1:
			_DecFlags = DTS_MONO;
			break;

		// 2 channels
		case 2:
			_DecFlags = DTS_STEREO;
			break;

		// 4 channels
		case 4:
			_DecFlags = DTS_2F2R;
			break;

		// 6 channels
		case 6:
			_DecFlags = DTS_3F2R;
			if (LFE)
				_DecFlags |= DTS_LFE;
			break;

		default:
			env->ThrowError("m2AudioDTSSource: invalid number of channels requested");
		}

		if (Downmix < ChannelCount)
		{
			ChannelCount = Downmix;
			DecFlags     = _DecFlags;
		}
	}

	Info.nchannels = ChannelCount;

	// Enable automatic level adjustment
	DecFlags |= DTS_ADJUST_LEVEL;
	Level     = 1;

	// Get sampling rate
	Info.audio_samples_per_second = Samplerate;

	// Get number of frames
	FrameCount = int((StreamLength - StreamOffset) / FrameLength);
	
	// Get number of samples
	SampleCount            = __int64(FrameCount) * __int64(SamplesPerFrame);
	Info.num_audio_samples = SampleCount;

	// Initialize framebuffer
	BufferSize  = Info.nchannels * SamplesPerFrame;
	Buffer      = new float[BufferSize];
	BufferStart = BufferEnd = -1;
}

// Destructor
m2AudioDTSSource::~m2AudioDTSSource()
{
	// Close file
	if (Stream)
		vfclose(Stream);

	// Free memory
	if (State)
		dts_free(State);

	if (Frame)
		delete[] Frame;

	if (Buffer)
		delete[] Buffer;
}

/*-----------------------------------------------------------------------------
	m2AudioDTSSource >> IClip interface
-----------------------------------------------------------------------------*/

// Return VideoInfo
const VideoInfo& __stdcall m2AudioDTSSource::GetVideoInfo()
{
	return Info;
}

// Return video frame
PVideoFrame __stdcall m2AudioDTSSource::GetFrame(int n, IScriptEnvironment *env)
{
	return NULL;
}

// Return video field parity
bool __stdcall m2AudioDTSSource::GetParity(int n)
{
	return false;
}

// Return audio samples
void __stdcall m2AudioDTSSource::GetAudio(void *buf, __int64 start, __int64 count, IScriptEnvironment *env)
{
	__int64 i;			    	// cycle counter
	int     _FrameIndex;		// index of the previous frame
	float   *Input, *Output;	// buffer pointers
	__int64  Left;			    // number of samples to be copied from the buffer
	
	// Check if the requested samples are not out of range
	if (!SampleCount || (start < 0) || (count < 1))
		return;

	if (start + count > SampleCount)
		count = SampleCount - start;
	
	// Check if we must read a frame
	if (start < BufferStart || start > BufferEnd)
	{
		// Calculate the index and the sample range of the required frame
		_FrameIndex = FrameIndex;
		FrameIndex  = int(start / SamplesPerFrame);
		BufferStart = __int64(FrameIndex) * __int64(SamplesPerFrame);
		BufferEnd   = BufferStart + SamplesPerFrame;

		// Seek to the frame, if required
		if (FrameIndex != _FrameIndex + 1)
		{
			vfseek(Stream, StreamOffset + __int64(FrameIndex) * __int64(FrameLength), SEEK_SET);

			// Reset decoder
			if (State)
				dts_free(State);
			State = dts_init(Accel);
		}

		// Read the frame
		if (!ReadFrame())
			env->ThrowError("m2AudioDTSSource: error in file \"%s\"", StreamName);
	}

	// Start decoding
	Input  = Buffer + int(start - BufferStart) * ChannelCount;
	Output = (float *)buf;
	Left   = SamplesPerFrame - (start - BufferStart);

	while (count)
	{
		// Copy samples
		if (count < Left)
			Left = count;
		count -= Left;

		for (i = 0; i < Left * ChannelCount; i++)
			*Output++ = *Input++;

		// Check if we must read a frame
		if (count)
		{
			// Read the frame
			FrameIndex++;
			if (!ReadFrame())
				env->ThrowError("m2AudioDTSSource: error in file \"%s\"", StreamName);

			// Reset input buffer pointer
			Input = Buffer;
			Left  = SamplesPerFrame;
		}
	}

	// Calculate the sample range of the current frame
	BufferStart = __int64(FrameIndex) * __int64(SamplesPerFrame);
	BufferEnd   = BufferStart + SamplesPerFrame;
}

// Set cache hints
void __stdcall m2AudioDTSSource::SetCacheHints(int cachehints, int frame_range)
{
}

/*-----------------------------------------------------------------------------
	m2AudioDTSSource >> Operations
-----------------------------------------------------------------------------*/

// Synchronize with the audio stream
int m2AudioDTSSource::Synchronize(int& Length, int& Flags, int& Samplerate, int& Bitrate)
{
	int Delta = 0;								// number of bytes out of synch
	int _Length;								// backup of Length
	unsigned char Buffer[FRAME_HEADER_LEN];		// buffer for a frameheader

	// Backup Length
	_Length = Length;

	// Try synchronization
	while (Delta < _Length - 1)
	{
		// Read header
		if (vfread(Stream, Buffer, FRAME_HEADER_LEN) != FRAME_HEADER_LEN)
			return -1;

		// Get information
		Length = dts_syncinfo(State, Buffer, &Flags, &Samplerate, &Bitrate, &Length);

		if (Length)
		{
			// Success
			vfseek(Stream, -FRAME_HEADER_LEN, SEEK_CUR);
			return (Delta + 1);
		}
		else
		{
			// Try again
			vfseek(Stream, 1 - FRAME_HEADER_LEN, SEEK_CUR);
			Delta++;
		}
	}

	// Failure
	Length = _Length;
	return 0;
}

// Reads the next frame from the stream and decodes it into the framebuffer
bool m2AudioDTSSource::ReadFrame()
{
	int Sync;							// synch info
	int Length;							// length of the current frame
	int _Flags, _Samplerate, _Bitrate;	// bit stream information

	// Read next frame from stream
	if (vfread(Stream, Frame, FrameLength) != FrameLength)
	{
		// Error at the end of the stream, mute frame
		EmptyFrame();
		return true;
	}

	// Get frame information
	Length = dts_syncinfo(State, Frame, &_Flags, &_Samplerate, &_Bitrate, &Length);

	// Check if we need synchronization
	if (!Length)
	{
		// Seek back in the stream
		vfseek(Stream, 1 - FrameLength, SEEK_CUR);

		// Try synchronization
		Length = FrameLength;
		Sync   = Synchronize(Length, _Flags, _Samplerate, _Bitrate);
		
		// Serious damage in the stream, mute frame
		if (!Sync || Sync < 0)
		{
			// Mute frame
			EmptyFrame();
			return true;
		}

		// Check if the frame has proper length and BSI
		if ((Length != FrameLength) || (_Flags != Flags) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

		// Read frame
		if (vfread(Stream, Frame, FrameLength) != FrameLength)
		{
			// Error at the end of the stream, mute frame
			EmptyFrame();
			return true;
		}
	}
	else
		// Check if the frame has proper length and BSI
		if ((Length != FrameLength) || (_Flags != Flags) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

	// Decode frame
	if (!DecodeFrame())
		EmptyFrame();

	return true;
}

// Decode frame from the raw buffer
bool m2AudioDTSSource::DecodeFrame()
{
	int    blk, i;			// cycle counters
	float *Input, *Output;	// buffer pointers
	
	// Start decoding the frame
	if (dts_frame(State, Frame, &DecFlags, &Level, 0) || (dts_blocks_num(State) != BlocksPerFrame))
		return false;

	// Dynamic range compression
	if(0==_drc)
		dts_dynrng(State, 0, 0);

	// The decoded samples will be in the following sequencial order:
	//   front center, front left, front right, rear left, rear right, LFE
	// If one of the channels is not present, it is skipped and the following
	// channels are shifted accordingly. We have to put the channels in
	// the proper interleaved order, based on the requested channel config.
	
	switch (ChannelCount)
	{
	// 1 channel
	case 1:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BlocksPerFrame; blk++)
		{
			// Decode block
			if (dts_block(State))
				return false;

			// Copy channel
			Input = dts_samples(State);

			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				*Output++ = *Input++;
		}
		break;

	// 2 channels: left, right
	case 2:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BlocksPerFrame; blk++)
		{
			// Decode block
			if (dts_block(State))
				return false;

			// Interleave channels
			Input = dts_samples(State);

			// left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
		break;

	// 4 channels: front left, front right, rear left, rear right
	case 4:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BlocksPerFrame; blk++)
		{
			// Decode block
			if (dts_block(State))
				return false;

			// Interleave channels
			Input = dts_samples(State);

			// front left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// front right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// rear left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 2] = *Input++;

			// rear right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 3] = *Input++;

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
		break;

	// 6 channels: front left, front right, front center, LFE, rear left, rear right
	case 6:
		// Decode each block
		Output = Buffer;
		for (blk = 0; blk < BlocksPerFrame; blk++)
		{
			// Decode block
			if (dts_block(State))
				return false;

			// Interleave channels
			Input = dts_samples(State);

			// front center
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 2] = *Input++;

			// front left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount] = *Input++;

			// front right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 1] = *Input++;

			// rear left
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 4] = *Input++;

			// rear right
			for (i = 0; i < SAMPLES_PER_BLOCK; i++)
				Output[i * ChannelCount + 5] = *Input++;

			// LFE
			if (LFE)
			{
				for (i = 0; i < SAMPLES_PER_BLOCK; i++)
					Output[i * ChannelCount + 3] = *Input++;
			}
			else
			{
				for (i = 0; i < SAMPLES_PER_BLOCK; i++)
					Output[i * ChannelCount + 3] = 0;
			}

			// Next block
			Output += SAMPLES_PER_BLOCK * ChannelCount;
		}
	}

	return true;
}

// Empty the framebuffer
void m2AudioDTSSource::EmptyFrame()
{
	for (int i = 0; i < SamplesPerFrame * ChannelCount; i++)
		Buffer[i] = 0;
}

/*-----------------------------------------------------------------------------
	AviSynth plugin initialization
-----------------------------------------------------------------------------*/
/*
extern "C" __declspec(dllexport) const char * __stdcall AvisynthPluginInit2(IScriptEnvironment *env)
{
	env->AddFunction("m2AudioDTSSource", "s[channels]i", m2AudioDTSSource::Create, 0);

	return "`m2audio_dts' m2Audio DTS Decoder Plugin";
}*/