/*=============================================================================
	Module Name:
			m2audio_lpcm.cpp

	Abstract:
			m2Audio Linear PCM decoder plugin for AviSynth 2.5

	License:
            GNU General Public License (GPL)

	Revision History:
			* Oct  3, 2004:  Created by Attila Afra
			* Jan  8, 2005:  Support for all formats
			* Feb 13, 2005:  Minor changes

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

// Includes
#include <windows.h>
#include "vfm.h"
#include "avisynth.h"

#include "AudioHeader.h"

/*-----------------------------------------------------------------------------
	Constants
-----------------------------------------------------------------------------*/

#define FRAME_LENGTH	25200			// length of a frame (raw)

// Nibble macros
#define HI_NIB(b)	((b >> 4) & 0x0f)
#define LO_NIB(b)	(b & 0x0f)

/*-----------------------------------------------------------------------------
	m2AudioLPCMSource: Input filter class
-----------------------------------------------------------------------------*/


/*=============================================================================
	m2AudioLPCMSource >> Implementation
=============================================================================*/

AVSValue __cdecl m2AudioLPCMSource::Create(AVSValue args, void *, IScriptEnvironment *env)
{
	return new m2AudioLPCMSource(args[0].AsString(), args[1].AsInt(), args[2].AsInt(), args[3].AsInt(), env);
}

/*-----------------------------------------------------------------------------
	m2AudioLPCMSource >> Initialization
-----------------------------------------------------------------------------*/

// Constructor
m2AudioLPCMSource::m2AudioLPCMSource(const char *FileName, int SampleRate, int SampleBits, int Channels, IScriptEnvironment *env)
{	
	// Initialize pointers
	Stream			= NULL;
	Frame			= 0;
	Buffer			= AuxBuffer1 = AuxBuffer2 = 0;
	bIsWavFile		= false;
	char szRIFF[5]	= "";

	// Empty information structure
	memset(&Info, 0, sizeof(VideoInfo));
	
	// Store filename
	StreamName = new char[strlen(FileName) + 1];
	strcpy(StreamName, FileName);
	
	// Open file
	if (!(Stream = vfopen(StreamName)))
		env->ThrowError("m2AudioLPCMSource: unable to open file \"%s\"", StreamName);

	vfread(Stream, szRIFF, 4);
	szRIFF[4] = 0;	// Null terminate

	if ( stricmp(szRIFF, "RIFF") == 0 )
	{
		bIsWavFile = true;
		vfseek(Stream, 44, SEEK_SET);
	}
	else
		vfseek(Stream, 0, SEEK_SET);	// Not Wav file so start back at the beginning

	StreamLength = vflen(Stream) + (bIsWavFile == true ? 44 : 0);

	// Set sample type
	InQuant = SampleBits;

	// Check
	if ( bIsWavFile && Channels != 2 )
	{
		env->ThrowError("m2AudioLPCMSource: Wav Files can only be 2 Channel currently for NicAudio. Rip the file as a raw LPCM to have more channels. Or bitch to Nic at Doom9 about it.");
	}

	switch (InQuant)
	{
	// 16 bit
	case 16:
		Quant = InQuant;
		Info.sample_type = SAMPLE_INT16;
		break;

	// 20 & 24 bit
	case 20:
	case 24:
	case 32:
		Quant = 32;
		Info.sample_type = SAMPLE_INT32;
		break;

	default:
		env->ThrowError("m2AudioLPCMSource: unsupported sample precision defined");
	}

	// Set channel count
	InChannelCount = Channels;

	if ((InChannelCount < 1) || (InChannelCount > 8))
		env->ThrowError("m2AudioLPCMSource: unsupported number of channels defined");

	switch (InChannelCount)
	{
	// 3 channels not supported, but we can take 2
	case 3:
		ChannelCount = 2;
		break;

	// 5 channels not supported, but we can create a dummy LFE channel to get 6
	// 7 channels not supported, but we can take 5 and create a dummy LFE channel to get 6
	// 8 channels not supported, but we can take 6
	case 5:
	case 7:
	case 8:
		ChannelCount = 6;
		break;

	default:
		ChannelCount = InChannelCount;
	}

	Info.nchannels = ChannelCount;

	// Set sampling rate
	Info.audio_samples_per_second = SampleRate;

	// Calculate number of samples per channel
	SampleCount = (StreamLength / (InChannelCount * InQuant)) * 8;
	Info.num_audio_samples = SampleCount;

	// Calculate number of samples per channel per frame
	SamplesPerFrame = (FRAME_LENGTH * 8) / (InChannelCount * InQuant);

	// Initialize buffers
	Frame = new unsigned char[FRAME_LENGTH];
	FrameIndex = -2;
	BufferStart = BufferEnd = -1;

	// We need axiliary buffer(s) because we cannot copy the samples directly from the raw buffer
	AuxBuffer1Size = (SamplesPerFrame * InChannelCount * Quant) / 8;
	AuxBuffer1     = new unsigned char[AuxBuffer1Size];

	if (ChannelCount != InChannelCount)
	{
		AuxBuffer2Size = (SamplesPerFrame * ChannelCount * Quant) / 8;
		AuxBuffer2     = new unsigned char[AuxBuffer2Size];
	}
}

// Destructor
m2AudioLPCMSource::~m2AudioLPCMSource()
{
	// Close file
	if (Stream)
		vfclose(Stream);

	// Free memory
	if (Frame)
		delete[] Frame;

	if (AuxBuffer1)
		delete[] AuxBuffer1;

	if (AuxBuffer2)
		delete[] AuxBuffer2;
}

/*-----------------------------------------------------------------------------
	m2AudioLPCMSource >> IClip interface
-----------------------------------------------------------------------------*/

// Return VideoInfo
const VideoInfo& __stdcall m2AudioLPCMSource::GetVideoInfo()
{
	return Info;
}

// Return video frame
PVideoFrame __stdcall m2AudioLPCMSource::GetFrame(int n, IScriptEnvironment *env)
{
	return NULL;
}

// Return video field parity
bool __stdcall m2AudioLPCMSource::GetParity(int n)
{
	return false;
}

// Return audio samples
void __stdcall m2AudioLPCMSource::GetAudio(void *buf, __int64 start, __int64 count, IScriptEnvironment *env)
{
	__int64         i;							// cycle counter
	int            _FrameIndex;					// index of the previous frame
	unsigned char *InputBuf, *Input, *Output;	// buffer pointers
	__int64        Left;						// number of samples to be copied from the buffer
	
	// Check if the requested samples are not out of range
	if ((start < 0) || (count < 1) || (start + count > SampleCount))
		return;
	
	// Check if we must read a frame
	if (start < BufferStart || start > BufferEnd)
	{
		// Calculate the index and the sample range of the required frame
		_FrameIndex = FrameIndex;
		FrameIndex  = int(start / SamplesPerFrame);
		BufferStart = __int64(FrameIndex) * SamplesPerFrame;
		BufferEnd   = BufferStart + SamplesPerFrame;

		// Seek to the frame, if required
		if (FrameIndex != _FrameIndex + 1)
			vfseek(Stream, __int64(FrameIndex) * FRAME_LENGTH, SEEK_SET);

		// Read the frame
		if (!ReadFrame())
			env->ThrowError("AC3Source: error in file \"%s\"", StreamName);
	}

	// Start decoding
	InputBuf = Buffer;
	Input = InputBuf + (int(start - BufferStart) * ChannelCount * Quant) / 8;

	Output = (unsigned char *)buf;
	Left = SamplesPerFrame - (start - BufferStart);

	while (count)
	{
		// Copy samples
		if (count < Left)
			Left = count;
		count -= Left;

		for (i = 0; i < (Left * ChannelCount * Quant) / 8; i++)
			*Output++ = *Input++;

		// Check if we must read a frame
		if (count)
		{
			// Read the frame
			FrameIndex++;
			if (!ReadFrame())
				env->ThrowError("AC3Source: error in file \"%s\"", StreamName);

			// Reset input buffer pointer
			Input = InputBuf;
			Left  = SamplesPerFrame;
		}
	}

	// Calculate the sample range of the current frame
	BufferStart = __int64(FrameIndex) * SamplesPerFrame;
	BufferEnd   = BufferStart         + SamplesPerFrame;
}

// Set cache hints
void __stdcall m2AudioLPCMSource::SetCacheHints(int cachehints, int frame_range)
{
}

/*-----------------------------------------------------------------------------
	m2AudioLPCMSource >> Operations
-----------------------------------------------------------------------------*/

// Reads the next frame from the stream and copies the samples into the buffer
bool m2AudioLPCMSource::ReadFrame()
{
	int            i, j;			// counters
	unsigned char *Input, *Output;	// input/output buffer pointers

	// Read next frame from stream
	if ((FrameLength = vfread(Stream, Frame, FRAME_LENGTH)) < 0)
		// Read error
		return false;

	// Calculate number of samples per channel per frame
	SamplesPerFrame = (FrameLength * 8) / (InChannelCount * InQuant);

	// Convert samples
	Input = Frame;
	Output = AuxBuffer1;

	switch (InQuant)
	{
	// 16 bit
	case 16:
		for (i = 0; i < FrameLength; i += 2)
		{
			// They are in the correct order
			if ( bIsWavFile )
			{
				Output[i]     = Input[i];
				Output[i + 1] = Input[i + 1];
			}
			else
			{
				Output[i]     = Input[i + 1];
				Output[i + 1] = Input[i];
			}
		}
		break;

	// 20 bit
	case 20:
		unsigned char Packet[5];	// packet containing 2 samples

		for (i = 0; i < FrameLength / (5 * InChannelCount); i++)
		{
			for (j = 0; j < InChannelCount; j++)
			{
				Packet[0] = (HI_NIB(Input[4 * InChannelCount + j]) << 4) + HI_NIB(Input[4 * j + 1]);
				Packet[1] = (LO_NIB(Input[4 * j + 1]) << 4) + HI_NIB(Input[4 * j + 0]);
				Packet[2] = (LO_NIB(Input[4 * j + 0]) << 4) + LO_NIB(Input[4 * InChannelCount + j]);
				Packet[3] = Input[4 * j + 3];
				Packet[4] = Input[4 * j + 2];

				Output[0] = 0;
				Output[1] = HI_NIB(Packet[0]) << 4;
				Output[2] = (LO_NIB(Packet[0]) << 4) + HI_NIB(Packet[1]);
				Output[3] = (LO_NIB(Packet[1]) << 4) + HI_NIB(Packet[2]);
				Output[4] = 0;
				Output[5] = LO_NIB(Packet[2]) << 4;
				Output[6] = Packet[3];
				Output[7] = Packet[4];

				Output += 8;
			}

			Input += 5 * InChannelCount;
		}
		break;

	// 24 bit
	case 24:
		for (i = 0; i < FrameLength / (6 * InChannelCount); i++)
		{
			for (j = 0; j < 2 * InChannelCount; j++)
			{
				Output[0] = 0;
				Output[1] = Input[4 * InChannelCount + j];
				Output[2] = Input[2 * j + 1];
				Output[3] = Input[2 * j]; 

				Output += 4;
			}

			Input += 6 * InChannelCount;
		}
		break;
	}

	Input = AuxBuffer1;
	// Now "Input" contains 16 or 32 bit samples

	// Check if we have to adjust the channel configuration
	if (AuxBuffer2)
	{
		Output = AuxBuffer2;

		switch (InChannelCount)
		{
		// 3 channels
		case 3:
			// Discard the last channel
			switch (Quant)
			{
			// 16 bit
			case 16:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output = *(long *)Input;	Output += 4; Input += 4;  // Lf, Rf

					Input += 2;
				}
				break;

			// 32 bit
			case 32:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output = *(long *)Input;	Output += 4; Input += 4;  // Lf
					*(long *)Output = *(long *)Input;	Output += 4; Input += 4;  // Rf

					Input += 4;
				}
				break;
			}
			break;

		// 5 channels
		case 5:
			// Add dummy LFE channel
			switch (Quant)
			{
			// 16 bit
			case 16:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf, Rf
					*(short *)Output = *(short *)Input;	Output += 2; Input += 2;  // C
					*(short *)Output = 0;				Output += 2;			  // LFE (dummy)
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls, Rs
				}
				break;

			// 32 bit
			case 32:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // C
					*(long *)Output  = 0;				Output += 4;			  // LFE (dummy)
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rs
				}
				break;
			}
			break;

		// 7 channels
		case 7:
			// Discard the last 2 channels & add dummy LFE channel
			switch (Quant)
			{
			// 16 bit
			case 16:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf, Rf
					*(short *)Output = *(short *)Input;	Output += 2; Input += 2;  // C
					*(short *)Output = 0;				Output += 2;			  // LFE (dummy)
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls, Rs

					Input += 4;
				}
				break;

			// 32 bit
			case 32:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // C
					*(long *)Output  = 0;				Output += 4;			  // LFE (dummy)
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rs

					Input += 8;
				}
				break;
			}
			break;

		// 8 channels
		case 8:
			// Discard the last 2 channels
			switch (Quant)
			{
			// 16 bit
			case 16:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf, Rf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // C, LFE
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls, Rs

					Input += 4;
				}
				break;

			// 32 bit
			case 32:
				for (i = 0; i < SamplesPerFrame; i++)
				{
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Lf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rf
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // C
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // LFE
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Ls
					*(long *)Output  = *(long *)Input;	Output += 4; Input += 4;  // Rs

					Input += 8;
				}
				break;
			}
			break;
		}

		Input = AuxBuffer2;
	}

	// Success
	Buffer = Input;
	return true;
}

/*-----------------------------------------------------------------------------
	AviSynth plugin initialization
-----------------------------------------------------------------------------*/
/*
extern "C" __declspec(dllexport) const char * __stdcall AvisynthPluginInit2(IScriptEnvironment *env)
{
	env->AddFunction("m2AudioLPCMSource", "s[samplerate]i[samplebits]i[channels]i", m2AudioLPCMSource::Create, 0);

	return "`m2audio_lpcm' m2Audio LPCM Decoder Plugin";
}*/