/*=============================================================================
	Module Name:
			m2audio_mpa.cpp

	Abstract:
			m2Audio MPEG decoder plugin for AviSynth 2.5

	License:
            GNU General Public License (GPL)

	Revision History:
			* Jul 19, 2004:  Created by Attila Afra
			* Jul 20, 2004:  DAI (DVD Audio Index) input
						  	 Major bugfixes, improved error tolerance
			* Sep 13, 2004:  DAI replaced with VFM
			* Nov  6, 2004:  Bugfixes
			* Feb 13, 2005:  Minor changes
			* Feb 27, 2005:  Advanced dummy audio generation

	Notes:
			* MPEG Layer III support is disabled because it is unstable.

------------------------------------------------------------------------------
	Copyright  2004-2005 Attila T. Afra. All rights reserved.
=============================================================================*/

#include <windows.h>
#include <math.h>
#include <string.h>
#include "vfm.h"
#include "avisynth.h"

#include "AudioHeader.h"

#define FRAME_HEADER_LEN		4			// length of a frame's header

/*=============================================================================
	m2AudioMPASource >> Implementation
=============================================================================*/

AVSValue __cdecl m2AudioMPASource::Create(AVSValue args, void *, IScriptEnvironment *env)
{
	return new m2AudioMPASource(args[0].AsString(), args[1].AsInt(0), env);
}

/*-----------------------------------------------------------------------------
	m2AudioMPASource >> Initialization
-----------------------------------------------------------------------------*/

// Constructor
m2AudioMPASource::m2AudioMPASource(const char *FileName, int Downmix, IScriptEnvironment *env)
{
	int Channels;	// channel count
	int Sync;		// synch info
	int _DecFlags;	// temporary flags
	
	// Initialize pointers
	Stream = 0;
	Frame  = 0;
	Buffer = 0;
	
	// Store filename
	StreamName = new char[strlen(FileName) + 1];
	strcpy(StreamName, FileName);
	
	// Open file
	if (!(Stream = vfopen(StreamName)))
		env->ThrowError("m2AudioMPASource: unable to open file \"%s\"", StreamName);

	StreamLength = vflen(Stream);

	// Check if the stream is empty
	if (!StreamLength)
	{
		// Dummy audio
		Info.sample_type              = SAMPLE_INT16;
		Info.nchannels                = Downmix ? Downmix : 2;
		Info.audio_samples_per_second = 44100;
		Info.num_audio_samples        = 0;

		SampleCount                   = 0;

		return;
	}

	// Empty information structure
	memset(&Info, 0, sizeof(VideoInfo));

	// Find first frame
	FrameLength = MAX_DETECT_ERROR;
	Sync        = Synchronize(FrameLength, Layer, ChannelCount, Samplerate, Bitrate);
	if (Sync <= 0)
		env->ThrowError("m2AudioMPASource: \"%s\" is not a valid MPEG audio file", StreamName);

	// MPEG Layer III does not work correctly, so it's currently disabled
	if (Layer == 3)
		env->ThrowError("m2AudioMPASource: MPEG Layer III is not yet supported");

	// In Layer II single channel mode the bitrate must not exceed 192 kbps
	if ((Layer == 2) && (ChannelCount == 1) && (Bitrate > 192))
		env->ThrowError("m2AudioMPASource: \"%s\" is not a standard MPEG audio file", StreamName);

	StreamOffset = (Sync - 1) % FrameLength;
	vfseek(Stream, StreamOffset, SEEK_SET);

	// Get padding
	if (Layer == 1)
		Padding = 1;
	else
		Padding = 4;

	// Initialize raw buffer
	Frame      = new unsigned char[FrameLength * 2];
	FrameIndex = -2;
	
	if (Layer == 1)
		SamplesPerFrame = SAMPLES_PER_FRAME_L1;
	else
		SamplesPerFrame = SAMPLES_PER_FRAME_L2_L3;

	// Set sample type
	Info.sample_type = SAMPLE_FLOAT;
	
	// Set channel count
	if (Downmix && (Downmix != 1) && (Downmix != 2))
		env->ThrowError("m2AudioMPASource: Invalid number of channels requested", StreamName);

	if (ChannelCount == 2 && Downmix == 1)
		Info.nchannels = 1;
	else
		Info.nchannels = ChannelCount;

	// Set sampling rate
	Info.audio_samples_per_second = Samplerate;

	// Get number of frames
	double _FrameCount, _FrameCountInt;

#define D(x) ((double)x)

	//_FrameCount = (( double(StreamLength - StreamOffset) / double(Bitrate / 8 * 1000) ) / double(SamplesPerFrame)) * double(Samplerate);
	_FrameCount = (( D(StreamLength - StreamOffset) / D(Bitrate / 8 * 1000) ) / D(SamplesPerFrame)) * D(Samplerate);

	if (modf(_FrameCount, &_FrameCountInt) <= 0.5)
		FrameCount = int(_FrameCountInt);
	else
		FrameCount = int(ceil(_FrameCount));

	// Get average frame length
	AvgFrameLength = double(StreamLength - StreamOffset) / double(FrameCount);
	
	// Get number of samples
	SampleCount            = __int64(FrameCount) * __int64(SamplesPerFrame);
	Info.num_audio_samples = SampleCount;

	// Initialize framebuffer
	BufferSize  = Info.nchannels * SamplesPerFrame;
	Buffer      = new float[BufferSize];
	BufferStart = BufferEnd = -1;

	// Initialize MAD stuff
	mad_stream_init(&MADStream);
	mad_frame_init(&MADFrame);
	mad_synth_init(&MADSynth);
}

// Destructor
m2AudioMPASource::~m2AudioMPASource()
{
	// Close file
	if (Stream)
		vfclose(Stream);

	// Finish MAD stuff
	mad_stream_finish(&MADStream);
	mad_frame_finish(&MADFrame);
	mad_synth_finish(&MADSynth);

	// Free memory
	if (Frame)
		delete[] Frame;

	if (Buffer)
		delete[] Buffer;
}

/*-----------------------------------------------------------------------------
	m2AudioMPASource >> IClip interface
-----------------------------------------------------------------------------*/

// Return VideoInfo
const VideoInfo& __stdcall m2AudioMPASource::GetVideoInfo()
{
	return Info;
}

// Return video frame
PVideoFrame __stdcall m2AudioMPASource::GetFrame(int n, IScriptEnvironment *env)
{
	return NULL;
}

// Return video field parity
bool __stdcall m2AudioMPASource::GetParity(int n)
{
	return false;
}

// Return audio samples
void __stdcall m2AudioMPASource::GetAudio(void *buf, __int64 start, __int64 count, IScriptEnvironment *env)
{
	__int64  i;						// cycle counter
	int      _FrameIndex;			// index of the previous frame
	float   *Input, *Output;		// buffer pointers
	__int64  Offset;				// offset in the file to seek
	__int64  Left;					// number of samples to be copied from the buffer
	
	// Check if the requested samples are not out of range
	if (!SampleCount || (start < 0) || (count < 1))
		return;

	if (start + count > SampleCount)
		count = SampleCount - start;
	
	// Check if we must read a frame
	if (start < BufferStart || start > BufferEnd)
	{
		// Calculate the index and the sample range of the required frame
		_FrameIndex = FrameIndex;
		FrameIndex  = int(start / SamplesPerFrame);
		BufferStart = __int64(FrameIndex) * __int64(SamplesPerFrame);
		BufferEnd   = BufferStart + SamplesPerFrame;

		// Seek to the frame, if required
		if (FrameIndex != _FrameIndex + 1)
		{
			Offset = StreamOffset + __int64(floor(double(FrameIndex) * AvgFrameLength));
			if (Offset >= Padding)
				Offset -= __int64(Padding);
			vfseek(Stream, Offset, SEEK_SET);

			// Reset decoder
			mad_stream_finish(&MADStream);
			mad_frame_finish(&MADFrame);
			mad_synth_finish(&MADSynth);
			mad_stream_init(&MADStream);
			mad_frame_init(&MADFrame);
			mad_synth_init(&MADSynth);
		}

		// Read the frame
		if (!ReadFrame())
			env->ThrowError("m2AudioMPASource: error in file \"%s\"", StreamName);
	}

	// Start decoding
	Input  = Buffer + int(start - BufferStart) * Info.nchannels;
	Output = (float *)buf;
	Left   = SamplesPerFrame - (start - BufferStart);

	while (count)
	{
		// Copy samples
		if (count < Left)
			Left = count;
		count -= Left;

		for (i = 0; i < Left * Info.nchannels; i++)
			*Output++ = *Input++;

		// Check if we must read a frame
		if (count)
		{
			// Read the frame
			FrameIndex++;
			if (!ReadFrame())
				env->ThrowError("m2AudioMPASource: error in file \"%s\"", StreamName);

			// Reset input buffer pointer
			Input = Buffer;
			Left  = SamplesPerFrame;
		}
	}

	// Calculate the sample range of the current frame
	BufferStart = __int64(FrameIndex) * __int64(SamplesPerFrame);
	BufferEnd   = BufferStart + SamplesPerFrame;
}

// Set cache hints
void __stdcall m2AudioMPASource::SetCacheHints(int cachehints, int frame_range)
{
}

/*-----------------------------------------------------------------------------
	m2AudioMPASource >> Operations
-----------------------------------------------------------------------------*/

// Synchronize with the audio stream
int m2AudioMPASource::Synchronize(int& Length, int& Layer, int& Channels, int& Samplerate, int& Bitrate)
{
	int Delta = 0;								// number of bytes out of synch
	int _Length;								// backup of Length
	unsigned char Buffer[FRAME_HEADER_LEN];		// buffer for a frameheader

	// Backup Length
	_Length = Length;

	// Try synchronization
	while (Delta < _Length - 1)
	{
		// Read header
		if (vfread(Stream, Buffer, FRAME_HEADER_LEN) != FRAME_HEADER_LEN)
			return -1;

		// Get information
		Length = SyncInfo(Buffer, Layer, Channels, Samplerate, Bitrate);

		if (Length)
		{
			// Success
			vfseek(Stream, -FRAME_HEADER_LEN, SEEK_CUR);
			return (Delta + 1);
		}
		else
		{
			// Try again
			vfseek(Stream, 1 - FRAME_HEADER_LEN, SEEK_CUR);
			Delta++;
		}
	}

	// Failure
	Length = _Length;
	return 0;
}

// Reads the next frame from the stream and decodes it into the framebuffer
bool m2AudioMPASource::ReadFrame()
{
	int Sync;										// synch info
	int Length;										// length of the current frame
	int _Layer, _Channels, _Samplerate, _Bitrate;	// bit stream information

	// Read next frame header from stream
	if (vfread(Stream, Frame, FRAME_HEADER_LEN) != FRAME_HEADER_LEN)
	{
		// Error at the end of the stream, mute frame
		EmptyFrame();
		return true;
	}

	// Get frame information
	Length = SyncInfo(Frame, _Layer, _Channels, _Samplerate, _Bitrate);

	// Check if we need synchronization
	if (!Length)
	{
		// Seek back in the stream
		vfseek(Stream, 1 - FRAME_HEADER_LEN, SEEK_CUR);

		// Try synchronization
		Length = FrameLength;
		Sync = Synchronize(Length, _Layer, _Channels, _Samplerate, _Bitrate);

		// Serious damage in the stream, mute frame
		if (!Sync || Sync < 0)
		{
			// Mute frame
			EmptyFrame();
			return true;
		}

		// Check if the frame has proper BSI
		if ((_Layer != Layer) || (_Channels != ChannelCount) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

		// Read entire frame
		FrameLength = Length;
		if (vfread(Stream, Frame, FrameLength) != FrameLength)
		{
			// Error at the end of the stream, mute frame
			EmptyFrame();
			return true;
		}
	}
	else
	{
		// Check if the frame has proper BSI
		if ((_Layer != Layer) || (_Channels != ChannelCount) || (_Samplerate != Samplerate))
			// Fatal error, unable to continue decoding
			return false;

		// Read only frame data
		FrameLength = Length;
		if (vfread(Stream, Frame + FRAME_HEADER_LEN, FrameLength - FRAME_HEADER_LEN) != (FrameLength - FRAME_HEADER_LEN))
		{
			// Error at the end of the stream, mute frame
			EmptyFrame();
			return true;
		}
	}

	// Decode frame
	if (!DecodeFrame())
		// Frame is damaged, mute it
		EmptyFrame();

	return true;
}

// Decode frame from the raw buffer
bool m2AudioMPASource::DecodeFrame()
{
	int          i;					// cycle counters
	mad_fixed_t *Input, *Input2;	// input buffer pointers
	float       *Output;			// output buffer pointer	

	// Prepare buffer for the decoder
	mad_stream_buffer(&MADStream, Frame, FrameLength + MAD_BUFFER_GUARD);

	// Decode the frame
	if (mad_frame_decode(&MADFrame, &MADStream) == -1)
		return false;

	mad_synth_frame(&MADSynth, &MADFrame);

	// Copy samples
	switch (ChannelCount)
	{
	// 1 channel
	case 1:
		Output = Buffer;

		// Copy channel
		Input = MADSynth.pcm.samples[0];
		for (i = 0; i < MADSynth.pcm.length; i++)
			*Output++ = ConvertSample(*Input++);

		break;

	// 2 channels: left, right
	case 2:
		Output = Buffer;

		// Downmix?
		if (Info.nchannels == 1)
		{
			// Average channels
			Input  = MADSynth.pcm.samples[0];
			Input2 = MADSynth.pcm.samples[1];
			for (i = 0; i < MADSynth.pcm.length; i++)
				*Output++ = (ConvertSample(*Input++) + ConvertSample(*Input2++)) / 2;
		}
		else
		{
			// left
			Input = MADSynth.pcm.samples[0];
			for (i = 0; i < MADSynth.pcm.length; i++)
				Output[i * 2] = ConvertSample(*Input++);

			// right
			Input = MADSynth.pcm.samples[1];
			for (i = 0; i < MADSynth.pcm.length; i++)
				Output[i * 2 + 1] = ConvertSample(*Input++);
		}
	}

	return true;
}

// Empty the framebuffer
void m2AudioMPASource::EmptyFrame()
{
	for (int i = 0; i < SamplesPerFrame * Info.nchannels; i++)
		Buffer[i] = 0;
}

/*-----------------------------------------------------------------------------
	m2AudioMPASource >> Utilities
-----------------------------------------------------------------------------*/

// Get BSI from the synch header
int m2AudioMPASource::SyncInfo(unsigned char *Buf, int& Layer, int& Channels, int& Samplerate, int& Bitrate)
{
	// Constants
	static const int Bitrates2[3][15] = {
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
	};

	static const int Bitrates[3][15] = {
		{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
		{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320}
	};

	static const long Samplerates[4] = {44100, 48000, 32000, 0};

	// Variables
	int Header;
	int IsMPEG2, IsPadded, BitrateIndex, SamplerateIndex, Mode;
	int FrameSize;


	// Get header
	Header = *((int *)Buf);

	// Check if the header is valid
	if ((Header & 0xf0ff) != 0xf0ff)
		return 0;
	
	// 00 for layer ("layer 4") is not valid
	if (!(Header & 0x00000600))
		return 0;

	// 1111 for bitrate is not valid
	if ((Header & 0x00F00000) == 0x00F00000)
		return 0;

	// 11 for sampling frequency is not valid
	if ((Header & 0x000C0000) == 0x000C0000)
		return 0;

	// Retreive information from the header
	IsMPEG2         = !(Header & 0x00000800);
	Layer           = 4 - (Header >> 9) & 3;
	BitrateIndex    = (Header >> 20) & 15;
	Bitrate         = (IsMPEG2 ? Bitrates2 : Bitrates)[Layer - 1][BitrateIndex];
	SamplerateIndex = (Header >> 18) & 3;
	Samplerate      = Samplerates[SamplerateIndex];
	if (IsMPEG2)
		Samplerate >>= 1;
	IsPadded        = !!(Header & 0x00020000);
	Mode            = (Header >> 30) & 3;
	if (Mode == 3)
		Channels = 1;
	else
		Channels = 2;

	// Compute the frame size
	if (Layer == 1)
	{
		FrameSize = 4 * (12000 * Bitrate / Samplerate);
		if (IsPadded)
			FrameSize += 4;
	}
	else
	{
		if (IsMPEG2)
			FrameSize = 72000 * Bitrate / Samplerate;
		else
			FrameSize = 144000 * Bitrate / Samplerate;

		if (IsPadded)
			FrameSize++;
	}

	// Return the frame size
	return FrameSize;
}

// Convert MAD's 32-bit sample format to float
float m2AudioMPASource::ConvertSample(mad_fixed_t Sample)
{
	return (float(Sample) / float(1L << MAD_F_FRACBITS));
}

/*-----------------------------------------------------------------------------
	AviSynth plugin initialization
-----------------------------------------------------------------------------*/
/*
extern "C" __declspec(dllexport) const char * __stdcall AvisynthPluginInit2(IScriptEnvironment *env)
{
	env->AddFunction("m2AudioMPASource", "s[channels]i", m2AudioMPASource::Create, 0);

	return "`m2audio_mpa' m2Audio MPEG Decoder Plugin";
}*/