/*=============================================================================
	Module Name:
			vfm.c

	Abstract:
			Virtual File Map I/O library

	License:
            GNU General Public License (GPL)

	Revision History:
			* Sep  1, 2004:  Created by Attila Afra
			* Sep 12, 2004:  Serious bug in I/O preparation fixed

------------------------------------------------------------------------------
	Copyright  2004 Attila T. Afra. All rights reserved.
=============================================================================*/

// Includes
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include "vfm.h"

// Nick: Simple VFM ;)
FILE *vfopen(const char *filename)
{
	return fopen(filename, "rb");
}

int vfclose(VFILE *stream)
{
	if ( stream )
		fclose(stream);

	return 0;
}

int vfread(VFILE *stream, void *buffer, unsigned int count)
{
	return fread(buffer, 1, count, stream);
}

__int64 vfseek(VFILE *stream, __int64 offset, int origin)
{
	return fseek(stream, offset, origin);
}

// Return the linear position in a virtual file
__int64 vftell(VFILE *stream)
{
	return ftell(stream);
}

// Check if we are at the end of a virtual file
int vfeof(VFILE *stream)
{
	return feof(stream);
}

// Return the length of a virtual file
__int64 vflen(VFILE *stream)
{
	__int64 nRet;
	long nCurPos = ftell(stream);

	fseek(stream, 0, SEEK_END);
	nRet = ftell(stream);
	fseek(stream, nCurPos, SEEK_SET);

	return nRet;
}

/*

// Constants
#define VFM_VERSION		10	   // current file format version
#define EOF_ID			"."	   // end-of-file identifier
#define MAP_ALLOC_STEP  1024   // map array allocation step size


// Map entry structure
typedef struct
{
	__int64 Offset;		 // linear offset in the virtual mapped file
	__int64 Length;		 // length of mapped block

	int		FileIndex;	 // index of true file
	__int64 FileOffset;  // offset in the true file
} _VMAPENTRY;

// Virtual file handle structure
typedef struct
{
	int			Version;	 // file format version
	int		   *Files;		 // array of real file handles
	int			FileCount;   // number of real files
	_VMAPENTRY *Map;		 // map array
	int			MapSize;	 // number of elements in the map array

	__int64     Length;		 // length of the virtual file
	__int64     Position;	 // current virtual position

	int			File;		 // current real file handle
	int			BlockIndex;	 // index of the current block
	__int64		BlockOffset; // offset in the current block
} _VFILE;

//---------------------------------------------------------------------------

// Open a virtual file
VFILE *vfopen(const char *filename)
{
	_VFILE  *file;
	FILE    *mapf;				// map file
	int      i, j;				// counters
	int      temp;				// temp variable
	char     Path[_MAX_PATH];	// file path
	int		 Next      = 0;		// index of the next element in the map array
	__int64  CurOffset = 0;		// current linear offset in the virtual mapped file
	int 	 Error     = 0;     // error flag
	__int64 *FileLengths;		// array of the real files' lengths (temporary)

	// Check filename
	if (!filename)
		return NULL;

	// Open map file
	if ((mapf = fopen(filename, "rt")) == NULL)
		return NULL;

	// Allocate memory for the handle structure
	if ((file = malloc(sizeof(_VFILE))) == NULL)
	{
		fclose(mapf);
		return NULL;
	}

	// Read header and number of real files
	if (((fscanf(mapf, "VirtualFileMap%d\n%d\n", &file->Version, &file->FileCount)) != 2) ||
		(file->Version > VFM_VERSION))
	{
		fclose(mapf);
		free(file);
		return NULL;
	}

	// Allocate memory for the array of file handles
	if ((file->Files = malloc(file->FileCount * sizeof(int))) == NULL)
	{
		fclose(mapf);
		free(file);
		return NULL;
	}

	if ((FileLengths = malloc(file->FileCount * sizeof(__int64))) == NULL)
	{
		fclose(mapf);
		free(file->Files);
		free(file);
		return NULL;
	}

	// Open files
	for (i = 0; i < file->FileCount; i++)
	{
		// Read filename
		fscanf(mapf, "%[^\n]\n", Path);

		// Open file
		if (((file->Files[i] = _open(Path, _O_BINARY | _O_RDONLY)) == -1) ||
			(FileLengths[i] = _filelengthi64(file->Files[i])) == -1)
		{
			// Error, close opened files so far
			fclose(mapf);

			for (j = 0; j < i; j++)
				_close(file->Files[j]);

			if (file->Files[i] != -1)
				_close(file->Files[i]);

			// Free memory
			free(FileLengths);
			free(file->Files);
			free(file);
			return NULL;
		}
	}

	// Initialize map
	file->MapSize = MAP_ALLOC_STEP;
	file->Map     = malloc(file->MapSize * sizeof(_VMAPENTRY));
	
	// Parse map
	for (; ;)
	{
		// Check if the map must be extended
		if (Next == file->MapSize)
		{
			file->MapSize += MAP_ALLOC_STEP;
			file->Map      = realloc(file->Map, file->MapSize * sizeof(_VMAPENTRY));
		}

		// Read next line
		if ((temp = fscanf(mapf, "%d %I64x %I64x\n", &file->Map[Next].FileIndex,
										             &file->Map[Next].FileOffset,
										             &file->Map[Next].Length)) != 3)
		{
			// Incomplete entry
			if (temp)
				Error = 1;

			break;
		}

		// Check if the entry is valid
		if ((file->Map[Next].FileIndex > (file->FileCount - 1)) ||
			((file->Map[Next].FileOffset + file->Map[Next].Length) > FileLengths[file->Map[Next].FileIndex]))
		{
			Error = 1;
			break;
		}

		// Process entry
		file->Map[Next].Offset = CurOffset;
		CurOffset += file->Map[Next].Length;
		Next++;
	}

	// Check for EOF ID
	if (Error || !fscanf(mapf, "%s", Path) || strcmp(Path, EOF_ID))
	{
		// Error, close files and free memory
		fclose(mapf);
		for (j = 0; j < file->FileCount; j++)
			_close(file->Files[j]);

		free(FileLengths);
		free(file->Files);
		free(file->Map);
		free(file);
		return NULL;
	}

	// Resize map
	file->MapSize = Next;
	file->Map     = realloc(file->Map, (file->MapSize + 1) * sizeof(_VMAPENTRY));

	// Add a dummy entry to the end of the map
	file->Map[Next].Offset = CurOffset;

	// Close map file
	fclose(mapf);

	// Free temporary data
	free(FileLengths);

	// Prepare for I/O
	file->Length   = CurOffset;
	file->Position = 0;

	file->BlockIndex  = 0;
	file->BlockOffset = 0;
	file->File        = file->Files[file->Map[file->BlockIndex].FileIndex];

	if (_lseeki64(file->File, file->Map[file->BlockIndex].FileOffset, SEEK_SET) == -1)
	{
		// Error, close files and free memory
		for (j = 0; j < file->FileCount; j++)
			_close(file->Files[j]);

		free(file->Files);
		free(file->Map);
		free(file);
		return NULL;
	}
	
	// Return pointer
	return file;
}

// Close a virtual file
int vfclose(VFILE *stream)
{
	_VFILE *file = stream;
	int     i;				// counter

	// Check pointer
	if (!file)
		return -1;

	// Close files
	for (i = 0; i < file->FileCount; i++)
		_close(file->Files[i]);
	
	// Free memory
	free(file->Files);
	free(file->Map);
	free(file);

	// Success
	return 0;
}

// Read from a virtual file
int vfread(VFILE *stream, void *buffer, unsigned int count)
{
	_VFILE *file = stream;
	__int64 Left = count;	// bytes left to be read
	__int64 CurCount;		// number of bytes to be read from the current block
	char   *out  = buffer;  // output buffer pointer

	if (!file || !buffer)
		return -1;

	if (!count)
		return 0;

	// Reading loop
	while (Left)
	{
		// Check if we cannot read enough bytes from the current block
		if ((CurCount = file->Map[file->BlockIndex].Length - file->BlockOffset) < Left)
		{
			// Read data
			if (_read(file->File, out, (int)CurCount) == -1)
				return -1;

			Left -= CurCount;

			// Adjust output buffer pointer
			out += CurCount;

			// Next block
			if (file->BlockIndex + 1 >= file->MapSize)
			{
				// EOF
				file->BlockOffset = file->Map[file->BlockIndex].Length;
				break;
			}

			file->File = file->Files[file->Map[++(file->BlockIndex)].FileIndex];

			if (_lseeki64(file->File, file->Map[file->BlockIndex].FileOffset, SEEK_SET) == -1)
				return -1;

			file->BlockOffset = 0;
		}
		else
		{
			// Read data
			if (_read(file->File, out, (int)Left) == -1)
				return -1;

			file->BlockOffset += Left;
			Left = 0;

			// Done
			break;
		}
	}

	// Adjust the linear file position
	Left = (__int64)count - Left;
	file->Position += Left;

	// Return the number of bytes read
	return (int)Left;
}

// Seek in a virtual file
__int64 vfseek(VFILE *stream, __int64 offset, int origin)
{
	_VFILE *file = stream;
	int Index;					// block index
	int Low, High, Mid;			// used by the modified binary search algo

	if (!file)
		return -1;

	// Calculate absolute offset
	switch (origin)
	{
	case SEEK_CUR:
		offset += file->Position;
		break;

	case SEEK_END:
		offset += file->Length;
		break;

	case SEEK_SET:
		break;

	default:
		return -1;
	}

	// Check if we can seek to that position
	if (offset > file->Length || offset < 0)
		return -1;

	// Do a special binary search to find the proper block
	Low  = 0;
	High = file->MapSize;

	for (; ;)
	{
		if (High == Low + 1)
		{
			Index = Low;
			break;
		}

		Mid = (Low + High) / 2;

		if (offset < file->Map[Mid].Offset)
			High = Mid;
		else
			if (offset > file->Map[Mid].Offset)
				Low = Mid;
			else
			{
				Index = Mid;
				break;
			}
	}

	// Seek to the proper block
	file->File        = file->Files[file->Map[Index].FileIndex];
	file->BlockIndex  = Index;
	file->BlockOffset = offset - file->Map[Index].Offset;

	if (_lseeki64(file->File, file->Map[Index].FileOffset + file->BlockOffset, SEEK_SET) == -1)
		return -1;
	
	file->Position = offset;
	return offset;
}

// Return the linear position in a virtual file
__int64 vftell(VFILE *stream)
{
	_VFILE *file = stream;

	if (file)
		return file->Position;
	else
		return -1;
}

// Check if we are at the end of a virtual file
int vfeof(VFILE *stream)
{
	_VFILE *file = stream;

	if (file)
		return file->Position >= file->Length;
	else
		return -1;
}

// Return the length of a virtual file
__int64 vflen(VFILE *stream)
{
	_VFILE *file = stream;

	if (file)
		return file->Length;
	else
		return -1;
}*/